//> using scala "2.13"
//> using lib "ch.unibas.cs.gravis::scalismo-ui:0.90.0"

// Basic geometric primitives
import scalismo.geometry.{_3D, Point, Point3D}
import scalismo.geometry.{EuclideanVector}
import scalismo.geometry.{IntVector, IntVector3D}
import scalismo.geometry.Landmark

import scalismo.common.PointId

// Geometric objects
import scalismo.mesh.TriangleMesh
import scalismo.mesh.TriangleId
import scalismo.image.{DiscreteImage, DiscreteImage3D}
import scalismo.statisticalmodel.PointDistributionModel

// IO Methods
import scalismo.io.ImageIO;
import scalismo.io.StatisticalModelIO
import scalismo.io.{MeshIO, StatisticalModelIO}

// Visualization
import scalismo.ui.api.ScalismoUI
import scalismo.ui.api.LandmarkView

import scalismo.io.LandmarkIO
import scalismo.registration.LandmarkRegistration
import scalismo.geometry._
import scalismo.common._
import scalismo.transformations._
import scalismo.io.MeshIO
import scalismo.common.interpolation._
import scalismo.common.interpolation.TriangleMeshInterpolator3D
import scalismo.ui.api._

object FYP extends App {

  scalismo.initialize()
  implicit val rng = scalismo.utils.Random(42)

  val ui = ScalismoUI()

  //Import skull meshes
  val meshFiles = new java.io.File("skulls/").listFiles.take(5)
  val dataset = meshFiles.map{ f => MeshIO.readMesh(f).get}
  val reference: TriangleMesh[_3D] = dataset(0)
  val skullGroup = ui.createGroup("skulls")
  val referenceView = ui.show(reference, "reference")
  (1 until 5).foreach{i :Int => ui.show(skullGroup, dataset(i), "skull"+i)}

  //Import landmarks
  import scalismo.io.LandmarkIO
  import scalismo.registration.LandmarkRegistration

  val lmGroup = ui.createGroup("landmarks")
  val reference_landmarks : Seq[Landmark[_3D]] = LandmarkIO.readLandmarksJson3D(new java.io.File("landmarks/reference.json")).get
  val reference_landmarksView = reference_landmarks.map(lm => ui.show(lmGroup, lm, s"${lm.id}"))
  val skull1_landmarks : Seq[Landmark[_3D]] = LandmarkIO.readLandmarksJson3D(new java.io.File("landmarks/skull1.json")).get
  val skull1_landmarksView = reference_landmarks.map(lm => ui.show(lmGroup, lm, s"${lm.id}"))
  val skull2_landmarks : Seq[Landmark[_3D]] = LandmarkIO.readLandmarksJson3D(new java.io.File("landmarks/skull2.json")).get
  val skull2_landmarksView = reference_landmarks.map(lm => ui.show(lmGroup, lm, s"${lm.id}"))
  val skull3_landmarks : Seq[Landmark[_3D]] = LandmarkIO.readLandmarksJson3D(new java.io.File("landmarks/skull3.json")).get
  val skull3_landmarksView = reference_landmarks.map(lm => ui.show(lmGroup, lm, s"${lm.id}"))
  val skull4_landmarks : Seq[Landmark[_3D]] = LandmarkIO.readLandmarksJson3D(new java.io.File("landmarks/skull4.json")).get
  val skull4_landmarksView = reference_landmarks.map(lm => ui.show(lmGroup, lm, s"${lm.id}"))
  
  //Rigid Alignment using landmarks
  import scalismo.registration.LandmarkRegistration
  val aligned_skullGroup = ui.createGroup("aligned skulls")

  val bestTransform1 : RigidTransformation[_3D] = LandmarkRegistration.rigid3DLandmarkRegistration(skull1_landmarks, reference_landmarks, center = Point(0, 0, 0))
  val aligned_skull1 = dataset(1).transform(bestTransform1)
  val aligned_skull1View = ui.show(aligned_skullGroup, aligned_skull1, "aligned skull1")

  val bestTransform2 : RigidTransformation[_3D] = LandmarkRegistration.rigid3DLandmarkRegistration(skull2_landmarks, reference_landmarks, center = Point(0, 0, 0))
  val aligned_skull2 = dataset(2).transform(bestTransform2)
  val aligned_skull2View = ui.show(aligned_skullGroup, aligned_skull2, "aligned skull2")

  val bestTransform3 : RigidTransformation[_3D] = LandmarkRegistration.rigid3DLandmarkRegistration(skull3_landmarks, reference_landmarks, center = Point(0, 0, 0))
  val aligned_skull3 = dataset(3).transform(bestTransform3)
  val aligned_skull3View = ui.show(aligned_skullGroup, aligned_skull3, "aligned skull3")

  val bestTransform4 : RigidTransformation[_3D] = LandmarkRegistration.rigid3DLandmarkRegistration(skull4_landmarks, reference_landmarks, center = Point(0, 0, 0))
  val aligned_skull4 = dataset(4).transform(bestTransform4)
  val aligned_skull4View = ui.show(aligned_skullGroup, aligned_skull4, "aligned skull4")

  val alignedMeshes : IndexedSeq[TriangleMesh[_3D]] = IndexedSeq(aligned_skull1, aligned_skull2, aligned_skull3, aligned_skull4)

  //Remove unnecessary components
  skullGroup.remove()
  lmGroup.remove()

  //Define mean and kernel (covariance function)
  import scalismo.io.{StatisticalModelIO, MeshIO}
  import scalismo.statisticalmodel._
  import scalismo.registration._
  import scalismo.statisticalmodel.dataset._
  import scalismo.numerics.PivotedCholesky.RelativeTolerance 
  import scalismo.kernels._
  import scalismo.numerics.UniformMeshSampler3D
  import breeze.linalg.{DenseMatrix, DenseVector}

  val referenceMesh = dataset(0)

  val modelGroup = ui.createGroup("gp-model")
  val referenceView1 = ui.show(modelGroup, referenceMesh, "reference")

  val zeroMean = Field(EuclideanSpace3D, (pt: Point[_3D]) => EuclideanVector3D(0, 0, 0))

  case class MatrixValuedGaussianKernel3D(sigma2: Double) extends MatrixValuedPDKernel[_3D]() {

    override def outputDim: Int = 3
    override def domain: Domain[_3D] = EuclideanSpace3D;

    override def k(x: Point[_3D], y: Point[_3D]): DenseMatrix[Double] = {
      DenseMatrix.eye[Double](outputDim) * Math.exp(-(x - y).norm2 / sigma2)
    }
  }

  val scalarValuedGaussianKernel: PDKernel[_3D] = GaussianKernel3D(sigma = 100.0)
  val matrixValuedGaussianKernel = DiagonalKernel3D(
    scalarValuedGaussianKernel,
    scalarValuedGaussianKernel,
    scalarValuedGaussianKernel
  )

  DiagonalKernel3D(scalarValuedGaussianKernel, 3)

  val gp = GaussianProcess3D[EuclideanVector[_3D]](zeroMean, matrixValuedGaussianKernel)

  val sampleGroup = ui.createGroup("samples")
  val sample = gp.sampleAtPoints(referenceMesh)
  ui.show(sampleGroup, sample, "gaussianKernelGP_sample")

  val interpolatedSample = sample.interpolate(TriangleMeshInterpolator3D())
  val deformedMesh = referenceMesh.transform((p: Point[_3D]) => p + interpolatedSample(p))
  ui.show(sampleGroup, deformedMesh, "deformed mesh")

  val lowRankGP = LowRankGaussianProcess.approximateGPCholesky(
    referenceMesh,
    gp,
    relativeTolerance = 0.01,
    interpolator = TriangleMeshInterpolator3D[EuclideanVector[_3D]]()
  )

  val defField: Field[_3D, EuclideanVector[_3D]] = lowRankGP.sample

  referenceMesh.transform((p: Point[_3D]) => p + defField(p))

  val pdm = PointDistributionModel3D(referenceMesh, lowRankGP)
  val pdmView = ui.show(modelGroup, pdm, "group")


}